//***************************************************************************
//  Genetic Algorithm Viewer 1.0
//    26/5/2000
//
//  This program uses a GA to find a Biomorph Genome
//    Jean-Philippe Rennard May 2000
//    alife@rennard.org
//    http://www.rennard.org/alife
//
//    This program is free software; you can redistribute it and/or
//    modify it under the terms of the GNU General Public License
//    as published by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//***************************************************************************

package Gav;

import java.awt.*;
import java.util.*;
/** CBiomorph : defines a Biomorph as a Biot */
class CBiomorph extends CBiot
{
  static final int MaxBranch=0x9;
  static final int MinBranch=0x3;
  static final int GeneBitLength=45;
  static final int NbGene=9;

  final int GenomeScale=0xF;
  final long DefaultGenome=0x15FFC0F7FFFFL;
  final long MaskGenome=0xFFFFFFFFFFFFFFFFL;
  final int Delta=GenomeScale/GenomeScale;
  // Used for BiomSize
  float XMax=Float.NEGATIVE_INFINITY;
  float XMin=Float.MAX_VALUE;
  float YMax=Float.NEGATIVE_INFINITY;
  float YMin=Float.MAX_VALUE;

  int XOffsets[];
  int YOffsets[];

  CGavUniverse MyUniverse;
  long BiomGenome;
  /** points Vector. Contains ending point of each branch.
  Represents Phenotype in GA */
  Vector BiomPhenotype;
  /** Position on Grid */
  int Place;
  double Fitness;
  /** Pointer to App GraphTools */
  JPGraphTools GT;

  /** Create a Biom in Universe */
  public CBiomorph(CGavUniverse univ) {
    BiomGenome=MaskGenome;
    MyUniverse=univ;
    BiomPhenotype=new Vector(NbGene);
    GT= MyUniverse.App.GrTools;
    XOffsets=new int[MyUniverse.NbDirections];
    YOffsets=new int[MyUniverse.NbDirections];
    Place=1;
    Direction = 2;  // Up
    Position.x=0;
    Position.y=0;
    SetDefaultGenes();
    GenPhenotype();
  }

  /** Randomly creates Biomorph at Place*/
  public CBiomorph(CGavUniverse univ,boolean rnd,int place)
  {
    BiomGenome=MaskGenome;
    MyUniverse=univ;
    BiomPhenotype=new Vector(NbGene);
    GT=MyUniverse.App.GrTools;
    XOffsets=new int[MyUniverse.NbDirections];
    YOffsets=new int[MyUniverse.NbDirections];
    Place=place+1;
    Direction= 2; // Up
    Position.x=0;
    Position.y=0;
    if(rnd) {
      for(int i=0;i<NbGene;i++)
        SetGene(i+1,(int)(Math.random()*(GenomeScale+1)) *
                  (Math.random()>0.5 ? Delta : -1*Delta));
      SetGene(9,(int)(Math.random()*(MaxBranch))+1);
    }
    else
      SetDefaultGenes();
    GenPhenotype();
  }

  /** Creates a descendant with a mutation on gene nummutant */
  public CBiomorph(CBiomorph father,int nummutant) {
    BiomGenome=MaskGenome;
    MyUniverse=father.MyUniverse;
    BiomPhenotype=new Vector(NbGene);
    GT=MyUniverse.App.GrTools;
    XOffsets=new int[MyUniverse.NbDirections];
    YOffsets=new int[MyUniverse.NbDirections];
    // In normal display - e.d. not random gen - the Place is linked to
    // the gen mutation.
    Place=nummutant+1;
    for(int i=0;i<NbGene;i++) {
      SetGene(i+1,father.GetGene(i+1));
    }
    Direction = father.Direction;
    Position.x=father.Position.x;
    Position.y=father.Position.y;
    int genecrt=(int)((nummutant-1)/2)%9+1;
    int d=Delta;
    if((nummutant)%2==1)
      d *= -1;
    SetGene(genecrt,father.GetGene(genecrt)+d);
    GenPhenotype();
  }

  /** Creates an offspring */
  public CBiomorph(CBiomorph father,CBiomorph mother,int place) {
    BiomGenome=MaskGenome;
    MyUniverse=father.MyUniverse;
    BiomPhenotype=new Vector(NbGene);
    GT=MyUniverse.App.GrTools;
    XOffsets=new int[MyUniverse.NbDirections];
    YOffsets=new int[MyUniverse.NbDirections];
    Place=place+1;

    // Doesn't cross branching depth
    int CrossPlace=(int)Math.round(Math.random()*(GeneBitLength-5));
    long bg=0x1L;
    bg <<= CrossPlace;
    bg -= 1;
    BiomGenome = father.BiomGenome & bg;
    bg = mother.BiomGenome & ~bg;
    BiomGenome |= bg;
    Direction = father.Direction;
    Position.x=father.Position.x;
    Position.y=father.Position.y;
    MutateBiom();
    GenPhenotype();
  }

  /** Mutate a Biom */
  void MutateBiom() {
    if(MyUniverse.MutateGene) {
      int mutgene=(int)(Math.random()*NbGene+1);
      if(Math.random()<MyUniverse.MutationRate) {
        SetGene(mutgene,GetGene(mutgene)+(Math.random()>0.5 ? 1 : -1));
      }
    }
    else {
      if(Math.random()<MyUniverse.MutationRate) {
        long bm=BiomGenome;
        long mask=1L << (long)(Math.random()*43);
        long r=BiomGenome & mask;
        if(r!=0)
          BiomGenome &= ~mask;
        else
          BiomGenome |= mask;
      }
    }
  }

  /** Set Biom genome to default values */
  void SetDefaultGenes() {
    BiomGenome=MaskGenome;
    SetGene(1, GenomeScale);
    SetGene(2, GenomeScale);
    SetGene(3, GenomeScale);
    SetGene(4, -1*GenomeScale);
    SetGene(5, -1*GenomeScale);
    SetGene(6, 0);
    SetGene(7, 1*GenomeScale);
    SetGene(8, 1*GenomeScale);
    SetGene(9, 5);
  }

  /** Gene setter */
	public void SetGene(int GeneNum, int GeneValue) {
    // Normalisation
    if(GeneNum!=9) {
      if(GeneValue>GenomeScale)
        GeneValue=GenomeScale;
      else if(GeneValue<GenomeScale*-1)
        GeneValue=GenomeScale*-1;
    }
      // Order Gene
    else {
      if(GeneValue>MaxBranch)
        GeneValue=MaxBranch;
      else if(GeneValue<MinBranch)
        GeneValue=MinBranch;
    }

    // Gene modification
    // creates FFFF 00000 FFFFF ...
    long bm=BiomGenome;
    long mask=0x1FL;
    mask <<= (GeneNum-1)*5;
    bm &= ~mask;
    mask=Math.abs(GeneValue);
    mask <<= (GeneNum-1)*5;
    if(GeneValue>0)
      mask |= (0x1L << ((GeneNum)*5)-1 );
    bm |= mask;
    mask=0x1L << this.GeneBitLength;
    mask -= 1;
    bm &= mask;
    BiomGenome=bm;
  }

  /** Return the nth Gene as Integer */
	int GetGene(int GeneNum) {
    long genecrt=BiomGenome >>> ((GeneNum-1)*5);
    return((int)( (genecrt & 0xFL) * ( (genecrt & 0x10L)>0 ? 1 : -1)));
  }

  /** Calculate X and Y Offsets from the genes */
  void CalcXOffset() {
    // Load up the X offsets from the genes
    XOffsets[0] = GetGene(2) * -1;
    XOffsets[1] = GetGene(1) * -1;
    XOffsets[2] = 0;
    XOffsets[3] = GetGene(1);
    XOffsets[4] = GetGene(2);
    XOffsets[5] = GetGene(3);
    XOffsets[6] = 0;
    XOffsets[7] = GetGene(3) * -1;

    // Load up the Y offsets from the genes
    YOffsets[0] = GetGene(6);
    YOffsets[1] = GetGene(5);
    YOffsets[2] = GetGene(4);
    YOffsets[3] = GetGene(5);
    YOffsets[4] = GetGene(6);
    YOffsets[5] = GetGene(7);
    YOffsets[6] = GetGene(8);
    YOffsets[7] = GetGene(7);
  }

  /** Live function : Draw */
  public void Live(JPView v) {
    CGavView bv=(CGavView) v;

    // How deep will we branch?
    int Order = GetGene(9);
    // Calculate X and Y Offset arrays
    CalcXOffset();

    int xOffs=0;
    int yOffs=0;
    float lh=0.0F;
    float lv=0.0F;
    Position.x=Position.y=0;
    XMax=Float.NEGATIVE_INFINITY;
    XMin=0.0F;
    YMax=Float.NEGATIVE_INFINITY;
    YMin=0.0F;

    CalcBiomSize(Position.x,Position.y,Order,Direction,XOffsets,YOffsets);
    Color crtcol;
    crtcol=bv.GrBuff.getColor();
    if( bv.GetType()==CGavView.TYPEMAIN) {
      if(Selected==1)
        v.GrBuff.setColor(Color.red);
      else if(Selected==2)
        v.GrBuff.setColor(Color.blue);
      xOffs= (int)( (((Place-1)%bv.NbHorizPlace))*bv.HorizStepD );
      yOffs= (int) ( ((int)((Place-1)/bv.NbHorizPlace+1)*bv.VerticStepD) );
    }
    else {
      xOffs=0;
      yOffs=v.Height;
    }
    lh=(float) 1.1*Math.abs(XMax-XMin);
    lv=(float) 1.1*(Math.abs(YMax-YMin));
    GT.Resize(bv.HorizStep,bv.VerticStep,lh,lv);
    GT.SetOrigOffset(xOffs,yOffs);
    Position.x=lh/2;
    Position.y=(YMin<0 ? Math.abs(YMin)+lv*0.05F : lv*0.05F);

    Draw(bv.GrBuff, Position.x, Position.y, Order, Direction, XOffsets, YOffsets);
    bv.GrBuff.setColor(crtcol);
  }

  /** Generate Vector Phenotype */
  void GenPhenotype() {
    CalcXOffset();
    int Length=NbGene-1;
    for(int i=Length;i>0;i--) {
      float NewX = 0.0F + i * XOffsets[(Direction-(Length-i)+ MyUniverse.NbDirections)%
                                              MyUniverse. NbDirections];
      float NewY = 0.0F - i * YOffsets[(Direction-(Length-i)+ MyUniverse.NbDirections)%
                                              MyUniverse. NbDirections];
      BiomPhenotype.addElement(new PosReal(NewX-0.0F,NewY-0.0F));
    }
    // Gene 9
    BiomPhenotype.addElement(new Integer(GetGene(9)));
  }

  void CalcBiomSize() {
    CalcXOffset();
    int Length=MaxBranch;
    float NewX=0;
    float NewY=0;
    for(int i=Length;i>0;i--) {
      NewX = NewX + i * XOffsets[(Direction-(Length-i)+ MyUniverse.NbDirections)%
                                              MyUniverse. NbDirections];
      NewY = NewY - i * YOffsets[(Direction-(Length-i)+ MyUniverse.NbDirections)%
                                              MyUniverse. NbDirections];
    }
    XMax=NewX*(float)1.1;
    XMin=(float)-1.0*XMax;
    YMax=NewY;
  }

  /** Determines the global size of the Biom */
  void CalcBiomSize(float X, float Y,int Length,int Direction, int XOffsets[], int YOffsets[])
  {
    Direction = (Direction + MyUniverse.NbDirections)%MyUniverse. NbDirections ;
    float NewX = X + Length * XOffsets[Direction];
    float NewY = Y - Length * YOffsets[Direction];
    if(NewX>XMax) XMax=NewX;
    if(NewX<XMin) XMin=NewX;
    if(NewY>YMax) YMax=NewY;

    if(YMax<0) YMax=0;

    if(NewY<YMin) YMin=NewY;

    if(Length>0) {
        // Draw one to the left
        CalcBiomSize(NewX, NewY, Length-1, Direction-1, XOffsets, YOffsets);
        // Draw one to the right
        CalcBiomSize(NewX, NewY, Length-1, Direction+1, XOffsets, YOffsets);
    }
  }

  /** Recursive Drawing algorithm */
  void Draw(Graphics g, float X, float Y, int Length, int Direction,
                                          int XOffsets[], int YOffsets[])
  {
    Direction = (Direction + MyUniverse.NbDirections)%MyUniverse. NbDirections ;
    float NewX = X + Length * XOffsets[Direction];
    float NewY = Y - Length * YOffsets[Direction];
    g.drawLine(GT.LToRX(X),GT.LToRY(Y),GT.LToRX(NewX),GT.LToRY(NewY));
    if(Length>0) {
        // Draw left
        Draw(g, NewX, NewY, Length-1, Direction-1, XOffsets, YOffsets);
        // Draw right
        Draw(g, NewX, NewY, Length-1, Direction+1, XOffsets, YOffsets);
    }
  }

  /** return Genome as String */
  public String DispGenome() {
    String s="";
    int g;
    for(int i=0;i<NbGene;i++) {
      g=GetGene(i+1);
      if(g>0) s += "+"; else s += "-";
      if(Math.abs(g)<10) s += "0";
      s += Math.abs(g);
      s += "/";
    }
    s += " -- "+Long.toBinaryString(BiomGenome);
    return(s);
  }
  /** Creates a binary string representation of genome */
  String GetBitGenomeString(long g) {
    String gs="";
    gs=Long.toBinaryString(g);
    return(gs);
  }

} // CBiomorph


