//***************************************************************************
//  Genetic Algorithm Viewer 1.0
//    26/5/2000
//
//  This program uses a GA to find a Biomorph Genome
//    Jean-Philippe Rennard May 2000
//    alife@rennard.org
//    http://www.rennard.org/alife
//
//    This program is free software; you can redistribute it and/or
//    modify it under the terms of the GNU General Public License
//    as published by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//***************************************************************************
package Gav;

import java.awt.*;
import java.util.*;

/** Defines GAV Universe */
class CGavUniverse extends CUniverse implements Runnable
{
  static final int SCALENONE=0;
  static final int SCALEWINDOW=1;
  static final int SCALEEXPO=2;
  static final int SCALELINEAR=3;
  static final int SCALELINEAR2=4;
  // Default parameters
  static final boolean DefaultElitistSel=true;
  static final boolean DefaultGenCalcType=false;
  static final boolean DefaultMutateGene=true;
  static final int DefaultScalingType=SCALELINEAR2;
  static final int DefaultSpeed=25;
  static final int DefaultGrid=3;
  /** Tab of population and corresponding mutation rate */
  static int TabPop[]={8,2,24,5,99,75,499,1000,899,10000,1599,0};
  static int DefaultPopIdx=1;

  /** The first bioms in Biots can be excluded from analysis.
  The #0 is the Current Researched Biom */
  final int BaseUsefulPop=1;
  /** Pointer to calling App */
  CGavApp App;
  /** View of the Universe */
  CGavView MainView;
  /** View of a genome */
  CGenView ViewGene1;
  CGenView ViewGene2;
  CGenView ViewGene3;
  /** Fitness View */
  CFitView ViewMaxFit;
  CFitView ViewAvgFit;
  /** Bit genome view */
  CGenBitView ViewGoalBitGen;
  CGenBitView ViewCurrentBitGen;
  // Fitness calc parameters
  /** Base for fitness calc. This is the maximum fitness Value */
  double BaseFitness;
  /** Max fitness in % */
  double PercentMaxFitness;
  /** Avg fitness in % */
  double PercentAvgFitness;
  /** Current Diversity*/
  double Diversity=0.0D;
  /** Some Bioms */
  CBiomorph CurrentBiom, BestBiom, WorseBiom, SelectedBiom;
  /** Elitist selection flag */
  boolean ElitistSel;
  /** Fitness calc type */
  boolean FitGenomeCalc;
  /** Mutation type */
  boolean MutateGene;
  /** Scaling type */
  int ScalingType;

  /** Construct World in Application */
  public CGavUniverse(CGavApp ga) {
    super((Container) ga);
    App=(CGavApp) MyApp;
    MainView=new CGavView(this,Color.black,CGavView.TYPEMAIN);
    ViewGene1=new CGenView(Color.black);
    ViewGene2=new CGenView(Color.black);
    ViewGene3=new CGenView(Color.black);
    ViewMaxFit=new CFitView(Color.black);
    ViewAvgFit=new CFitView(Color.black);
    ViewGoalBitGen=new CGenBitView(Color.red);
    ViewCurrentBitGen=new CGenBitView(Color.red);
    ScalingType=DefaultScalingType;
    InitDefaultWorld();
  }

   /** Default drawing */
   synchronized public void DrawWorld() {
    DrawWorld(MainView);
   }

  /** Draw World */
   public void DrawWorld(JPView jpv) {
    CGavView bv=(CGavView)jpv;
    bv.PaintBackground();
    int maxpop=Math.min(MainView.GetVisiblePop(),InitPopSize);
    for(int i=0;i<maxpop;i++) {
      ((CBiot)(Biots.elementAt(i))).Live(bv);
    }
    bv.repaint();
    DispStats();
  }

  /** View getter */
  public CGavView GetMainView() {
    return(MainView);
  }

  /** Init View */
  public void InitView() {
    MainView.ResizeGrBuff();
    MainView.PaintBackground();
    ViewGene1.ResizeGrBuff();
    ViewGene2.ResizeGrBuff();
    ViewGene3.ResizeGrBuff();
    ViewMaxFit.ResizeGrBuff();
    ViewAvgFit.ResizeGrBuff();
    ViewGoalBitGen.ResizeGrBuff();
    ViewCurrentBitGen.ResizeGrBuff();
  }
  /** Remove views */
  public void RemoveViews() {
    MainView.destroy();
    ViewGene1.destroy();
    ViewGene2.destroy();
    ViewGene3.destroy();
    ViewMaxFit.destroy();
    ViewAvgFit.destroy();
    ViewGoalBitGen.destroy();
    ViewCurrentBitGen.destroy();
  }
  /** Return current biom */
  public CBiomorph GetCurrentBiom() {
     return(CurrentBiom);
  }

  /** set Current Biom */
  public void SetCurrentBiom(int num) {
    if(CurrentBiom!=null)
      CurrentBiom.Selected=0;
//    CurrentBiom=((CBiomorph)(Biots.elementAt(num)));
    CurrentBiom=((CBiomorph)GetBiotAt(num));
    CurrentBiom.Selected=1;
    WorseBiom=CurrentBiom;
    BestBiom=CurrentBiom;
  }
  /** set selected Biom */
  public void SetSelectedBiom(int num) {
    if(SelectedBiom!=null)
      SelectedBiom.Selected=0;
    SelectedBiom=((CBiomorph)(GetBiotAt(num)));
    SelectedBiom.Selected=2;
  }

  /** Generate Random Bioms */
  void GenRndBioms() {
    Biots.removeAllElements();
    for(int i=0;i<=InitPopSize;i++) {
        Biots.addElement(new CBiomorph(this,true,i));
    }
    SetCurrentBiom(0);
    SelectedBiom=CurrentBiom;
  }

  /** Run the World in View e.g. creates a new pop and draw world. */
  public void Go(JPView jpv) {
    GenNewPop();
    Generation++;
    DrawWorld();
    if(SolFound) stop();
  }

  /** Default Run */
  public void Go() {
    Go(MainView);
  }

  // GA *********************************************************
  public boolean IsInit=false;
  public boolean IsRunning=false;
  final int MinPopSize=8;
  final int MaxPopSize=1000;
  int InitPopSize;
  double MutationRate;
  Thread GavUnivThread;

  /** Stats data */
  double GlobalFitness;
  double MaxFitness;
  double MinFitness;
  double AvgFitness;
  /** Fitness transfo data */
  double FtA;
  double FtB;
  /** Best Morph Index */
  int BestMorphIdx=0;
  /** Worse Morph Index */
  int WorseMorphIdx=0;
  /** Generation # */
  long Generation=0;
  /** Iteration delay */
  int Speed;
  /** Display each DispSpeed generations */
  int DispSpeed=1;
  /** Solution found */
  public boolean SolFound=false;

  /** Launch the Genetic Research. */
  void LaunchResearch() {
      if(GavUnivThread==null)
        GavUnivThread=new Thread(this);
      GavUnivThread.start();
  }
  /** Init the research process */
  void InitResearch(int mr) {
    // parameters normalisation
    if(mr<=0.0)
      MutationRate=0.0;
    else
      MutationRate=1.0/mr;
    Biots=null;
    Biots=new Vector();
    CurrentBiom.Place=1;
    Biots.addElement(CurrentBiom);
    GenInitPop();
    IsInit=true;
  }

  /** Main running loop */
  public void run() {
    Thread thisThread=Thread.currentThread();
    IsRunning=true;
    while(GavUnivThread==thisThread) {
     Go();
     try {
        Thread.sleep(Speed);
      } catch (InterruptedException e) { }
    } // while
  }

  public void stop() {
    IsRunning=false;
    GavUnivThread=null;
    App.stop();
  }
  /** This is the end... */
  public void end() {
     stop();
     RemoveViews();
  }
  
  /** Create a random init population */
  synchronized public void GenInitPop() {
    Generation=0;
    do {
      Biots.setSize(0);
      Biots.addElement(CurrentBiom);
      for(int i=1;i<InitPopSize;i++)
        Biots.addElement(new CBiomorph(this,true,i));
      CalcFitness();
    } while(AvgFitness<(BaseFitness/1000.0));  // avoid really bad pop
    ViewMaxFit.ClearBkg();
    ViewAvgFit.ClearBkg();
    DrawWorld();
  }

  /** Displays statistics */
  void DispStats() {
    if(!IsInit) {
      ViewMaxFit.DrawFitness(0);
      ViewAvgFit.DrawFitness(0);
      ViewGoalBitGen.DrawBitGenome(CurrentBiom);
      ViewCurrentBitGen.ClearBkg();
      ViewCurrentBitGen.repaint();
      App.FitBestGene.setText(App.Texts.Txt[27]+0);
      App.FitAvgGene.setText(App.Texts.Txt[28]+0);
      ViewGene1.DrawGenome(CurrentBiom);
      ViewGene2.DrawGenome(CurrentBiom);
      ViewGene3.DrawGenome(CurrentBiom);
      return;
    }
    ViewGene1.DrawGenome(CurrentBiom);
    ViewGene2.DrawGenome(SelectedBiom);
    ViewGene3.DrawGenome(WorseBiom);
    ViewMaxFit.DrawFitness(PercentMaxFitness);
    ViewAvgFit.DrawFitness(PercentAvgFitness);
    ViewGoalBitGen.DrawBitGenome(CurrentBiom);
    ViewCurrentBitGen.DrawBitGenome(SelectedBiom);
    String s;
    s=Double.toString(PercentMaxFitness);
    this.App.FitBestGene.setText(App.Texts.Txt[27]+s.substring(0,Math.min(10,s.length())));
    s=Double.toString(PercentAvgFitness);
    this.App.FitAvgGene.setText(App.Texts.Txt[28]+s.substring(0,Math.min(10,s.length())));
    this.App.Gen.setText(""+Generation);
  }

  /** Generate a new population according to fitness */
  synchronized void GenNewPop() {
    CBiomorph Father=null;
    CBiomorph Mother=null;
    Vector NewBiom=new Vector();
    int basesel=0;
    if(ElitistSel==true) {
    // Preserve the best Biom
      NewBiom.addElement(Biots.elementAt(BestMorphIdx));
      CalcBiomFitness((CBiomorph)(NewBiom.elementAt(0)));
      basesel=1;
    }
    for(int i=basesel;i<InitPopSize;i++) {
      Father=SelectBiom(GlobalFitness);
      Mother=SelectBiom(GlobalFitness);
      NewBiom.addElement(new CBiomorph(Father,Mother,i+1));
      CalcBiomFitness((CBiomorph)(NewBiom.elementAt(i)));
    }
    Sort(NewBiom);
    Biots.removeAllElements();
    Biots.addElement(CurrentBiom);
    int placecrt=2;
    for(int i=NewBiom.size()-1;i>=0;i--) {
      ((CBiomorph)(NewBiom.elementAt(i))).Place=placecrt++;
      Biots.addElement(NewBiom.elementAt(i));
    }
    CalcFitness();
  }

  /** Random Wheel selection */
  CBiomorph SelectBiom(double ttfitness) {
    double partsum=0.0;
    int j=BaseUsefulPop;// 0;
    double rnd=Math.random()*ttfitness;
    do {
      partsum += ((CBiomorph)GetBiotAt(j)).Fitness;
      j++;
    } while (partsum<rnd && j!= InitPopSize-1);
    return( (CBiomorph)GetBiotAt(j) );
  }
  /** Calculate pop fitness stats*/
  void CalcFitnessStats() {
    double gf=0;
    double max=Double.NEGATIVE_INFINITY;
    double min=Double.MAX_VALUE;
    double fitcrt;
    Diversity=0.0;
    for(int i=BaseUsefulPop;i<InitPopSize;i++) {
      fitcrt= ((CBiomorph)GetBiotAt(i)).Fitness;
      gf += fitcrt;
      Diversity += fitcrt-(gf/i);
      if(fitcrt>max) { max=fitcrt; BestMorphIdx=i; }
      if(fitcrt<min) { min=fitcrt; WorseMorphIdx=i; }
    }
    BestBiom=(CBiomorph)(GetBiotAt(BestMorphIdx));
    WorseBiom=(CBiomorph)(GetBiotAt(WorseMorphIdx));
    GlobalFitness=gf;
    AvgFitness=GlobalFitness/InitPopSize;
    MinFitness=min;
    MaxFitness=max;
    SelectedBiom=BestBiom;
  }

  /** Determine Fitness of each Biom */
  void CalcFitness() {
    for(int i=BaseUsefulPop;i<InitPopSize;i++)
      CalcBiomFitness((CBiomorph)GetBiotAt(i));
   ScalePop();
  }

  /** Calculate a Biom fitness */
  void CalcBiomFitness(CBiomorph biom) {
    double dx=0;
    double dy=0;

    // Genome Fitness calc type
    if(FitGenomeCalc) {
      BaseFitness=8*Math.pow(2.0*CurrentBiom.GenomeScale,2)+Math.pow(CurrentBiom.MaxBranch,3);
      for(int i=0;i<biom.NbGene;i++)  {
        dx = biom.GetGene(i+1)-CurrentBiom.GetGene(i+1);
        if (i==8) dx = Math.pow(dx,3);
        dx *= dx;
        dy += dx;
      }
      if(dy==0) SolFound=true;
      biom.Fitness=BaseFitness-dy;
    }

    // Phenotype Fitness calc type
    else {
      BaseFitness=0;
      for(int i=CBiomorph.NbGene-1;i>0;i--) {
        BaseFitness += i*i;
      }
      BaseFitness *= 1024;
      BaseFitness *= 2;
      BaseFitness += 1024*25*2;
      BaseFitness = Math.sqrt(BaseFitness);
      biom.Fitness=0.0;
      for(int i=0;i<CBiomorph.NbGene-1;i++) {
        dx=((PosReal)(biom.BiomPhenotype.elementAt(i))).x-
              ((PosReal)(CurrentBiom.BiomPhenotype.elementAt(i))).x;
        dx *= dx;
        dy=((PosReal)(biom.BiomPhenotype.elementAt(i))).y-
              ((PosReal)(CurrentBiom.BiomPhenotype.elementAt(i))).y;
        dy *= dy;
       biom.Fitness += (dx+dy);
      }
      Integer z;

      // gene 9 (depth)
      if(((Integer)(CurrentBiom.BiomPhenotype.elementAt(CurrentBiom.NbGene-1))).intValue() !=
              ((Integer)(biom.BiomPhenotype.elementAt(biom.NbGene-1))).intValue() )
          biom.Fitness += 1024*25*2;
      if(biom.Fitness==0.0)
        SolFound=true;
      biom.Fitness=Math.sqrt(biom.Fitness);
      double bm=biom.Fitness;
      biom.Fitness=BaseFitness-biom.Fitness;
    }
  }
  /** Transform pop fitness */
  void ScalePop() {
    CalcFitnessStats();
    double d=Diversity;
    PercentMaxFitness=MaxFitness/BaseFitness;
    PercentAvgFitness=AvgFitness/BaseFitness;
    switch(ScalingType) {
      case SCALENONE:
        break;
      case SCALEWINDOW:
        for(int i=BaseUsefulPop;i<InitPopSize;i++)
          ((CBiomorph)(GetBiotAt(i))).Fitness -= MinFitness;
        break;
      case SCALEEXPO:
        for(int i=BaseUsefulPop;i<InitPopSize;i++)
          ((CBiomorph)(GetBiotAt(i))).Fitness =
                Math.sqrt(((CBiomorph)(GetBiotAt(i))).Fitness+1.0);
      break;
      case SCALELINEAR:
        CalcTransfoFactor();
        double fcrt;
        for(int i=BaseUsefulPop;i<InitPopSize;i++) {
          fcrt=((CBiomorph)(GetBiotAt(i))).Fitness;
          ((CBiomorph)(GetBiotAt(i))).Fitness=Scale(fcrt,FtA,FtB);
        }
        break;
      case SCALELINEAR2:
        for(int i=BaseUsefulPop;i<InitPopSize;i++) {
          ((CBiomorph)(GetBiotAt(i))).Fitness=
                    Math.max(InitPopSize*10-((i-BaseUsefulPop)*10),10);
        }
      break;
    }
    CalcFitnessStats();
    Diversity=d;
  }

  /** Returns transfo fitness */
  double Scale(double f,double a,double b) {
    return(a*f+b);
  }
  /** Calculate linear transfo factor */
  void CalcTransfoFactor() {
    double fmul=2.0;
    double delta=0.0;
    if(MinFitness>(fmul*AvgFitness-MaxFitness)/(fmul-1.0)) {
      delta=MaxFitness-AvgFitness;
      FtA=(fmul-1.0)*AvgFitness/delta;
      FtB=AvgFitness*(MaxFitness-fmul*AvgFitness)/delta;
    }
    else {
      delta=AvgFitness-MinFitness;
      FtA=AvgFitness/delta;
      FtB=-1.0*MinFitness*AvgFitness/delta;
    }
  }

  /** Init a World */
  void InitWorld() {
    GavUnivThread=null;
    IsRunning=false;
    IsInit=false;
    SolFound=false;

    GenRndBioms();

  }
  /** Set default param */
  void InitDefaultWorld() {
    SetElitistSel(DefaultElitistSel);
    SetGenCalcType(DefaultGenCalcType);
    SetMutationType(DefaultMutateGene);
    SetScalingType(DefaultScalingType);
    SetSpeed(DefaultSpeed);
    SetPopSize(CGavUniverse.TabPop[CGavUniverse.DefaultPopIdx*2]+1);
    MutationRate=CGavUniverse.TabPop[CGavUniverse.DefaultPopIdx*2+1];
    InitWorld();
  }

  /** Set Elitist selection flag */
  public void SetElitistSel(boolean set) {
    ElitistSel=set;
  }
  /** Set genome type calc type flag */
  public void SetGenCalcType(boolean gen) {
    FitGenomeCalc=gen;
  }
  /** Set Mutation type */
  public void SetMutationType(boolean mutgene) {
    MutateGene=mutgene;
  }
  /** Get Elitist selection flag */
  public boolean GetElitistSel() {
    return(ElitistSel);
  }
  /** Set Caling type */
  public void SetScalingType(int st) {
    ScalingType=st;
  }
  /** Set current speed */
  public void SetSpeed(int s) {
    Speed=s;
  }
  /** Set Basic parameters */
  public void SetBasicParam(int pop,int mr) {
    InitPopSize=pop;
    MutationRate=mr;
  }
  /** Set a new population size with the corresponding grid. */
  public void SetPopSize(int pop) {
    InitPopSize=pop;
    switch(pop) {
      case 9:
        MainView.SetGrid(2);
        break;
      case 25:
        MainView.SetGrid(3);
        break;
      case 100:
        MainView.SetGrid(4);
        break;
      case 500:
        MainView.SetGrid(5);
        break;
      case 900:
        MainView.SetGrid(6);
        break;
      default:
        MainView.SetGrid(3);
    }
  }
  /** In/decrease grid */
  public void SetGrid(int incr) {
    MainView.SetGrid(incr);
  }

  /** Quick sort of a Morph Vector according to fitness */
  void Sort(Vector v) {
    v.trimToSize();
    QuickSort(v,0,v.size()-1);
  }
  /** Swap element i and j in Vector v */
  private void Swap(Vector v,int i, int j) {
    Object o=v.elementAt(i);
    v.setElementAt(v.elementAt(j),i);
    v.setElementAt(o,j);
  }
  /** Quick sort algorithm */
  void QuickSort(Vector v,int low, int up) {
    int lo = low;
    int hi = up;
    double mid;

    if ( up > low) {
      mid = ((CBiomorph)(v.elementAt((low+up)/2))).Fitness;
      while( lo <= hi ) {
	      while( ( lo < up ) && ( ((CBiomorph)(v.elementAt(lo))).Fitness < mid ))
          ++lo;
	      while( ( hi > low ) && ( ((CBiomorph)(v.elementAt(hi))).Fitness > mid ))
    		  --hi;
        if( lo <= hi ) {
          Swap(v, lo, hi);
          ++lo;
          --hi;
        }
      }
      if( low < hi )
        QuickSort( v, low, hi );
      if( lo < up )
        QuickSort( v, lo, up );
    }
  }
  
}