//***************************************************************************
//  Genetic Algorithm Viewer 1.0
//    26/5/2000
//
//  This program uses a GA to find a Biomorph Genome
//    Jean-Philippe Rennard May 2000
//    alife@rennard.org
//    http://www.rennard.org/alife
//
//    This program is free software; you can redistribute it and/or
//    modify it under the terms of the GNU General Public License
//    as published by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//***************************************************************************
package Gav;

import java.awt.*;
import java.awt.event.*;

//***************************************************************************
//      Views Def
//***************************************************************************
/** Gav Views : JPView extension to manage GAV **/
class CGavView extends JPView
{
  static final int MAXGRID=7;
  static final int TYPEMAIN=0;
  static final int TYPEZOOM=1;


  /** Pointer to the Universe **/
  CGavUniverse MyUniverse;
  int NbHorizPlace;
  int NbVerticPlace;
  int HorizStep, VerticStep;
  double HorizStepD, VerticStepD;
  int CurrentGrid;
  int VisiblePop=0;

  int Type;
  int SelectedGrid=0;
  int SelectedGene=0;
  /** Construct a GavView with background color and type */
  public CGavView(CGavUniverse univ,Color c, int type) {
      super(c);
      Type=type;
      MyUniverse=univ;
      NbHorizPlace=1;
      NbVerticPlace=1;
      SetGrid(MyUniverse.DefaultGrid);
  }

  /** Paint View Background */
  void PaintBackground() {
    this.ClearBkg();
    if(Type==TYPEMAIN) {
      Color colcrt=GrBuff.getColor();
      GrBuff.setColor(getBackground());
      GrBuff.fillRect(0,0,Width,Height);
      GrBuff.setColor(colcrt);
      GrBuff.drawRect(0,Width,Height,0);
      for(int i=1;i<NbHorizPlace;i++)
        GrBuff.drawLine((int)(i*(HorizStepD)),0,(int)(i*(HorizStepD)),Height);
      for(int i=1;i<NbVerticPlace;i++)
        GrBuff.drawLine(0,(int)(i*(VerticStepD)),Width,(int)(i*(VerticStepD)));
    }
  }

  /** Resize Graphic Buffer and Steps */
  public void ResizeGrBuff() {
    super.ResizeGrBuff();
    SetSteps();
  }
  /** Set the grid steps */
  void SetSteps() {
    HorizStep=Width/NbHorizPlace;
    VerticStep=Height/NbVerticPlace;
    HorizStepD=(double)Width/NbHorizPlace;
    VerticStepD=(double)Height/NbVerticPlace;
  }
  /** Type getter **/
  public int GetType() {
    return(Type);
  }

  /** Returns the # of the grid at point x,y **/
  int GetGridAt(int x,int y) {
    return( (int)(y/VerticStepD)*NbHorizPlace + (int)(x/HorizStepD)+1 );
  }

  /** Mouse Down : Process Selected grid and Gene/dir */
  public boolean mouseDown(Event e,int x,int y) {
    SelectedGrid=GetGridAt(x,y);
    if(SelectedGrid>MyUniverse.InitPopSize || MyUniverse.IsRunning)
      return(false);
//    System.out.println(((CBiomorph)(MyUniverse.Biots.elementAt(SelectedGrid-1))).DispGenome());
    CBiomorph b;
    // Zoom
    if(e.controlDown()) {
      Type=TYPEZOOM;
      MyUniverse.App.DisableRun();
      HorizStep=Width;
      HorizStepD=(double)Width;
      VerticStep=Height;
      VerticStepD=(double)Height;
      this.ClearBkg();
      b=((CBiomorph)(MyUniverse.GetBiotAt(SelectedGrid-1)));
      b.Live(this);
      MyUniverse.App.FitAvgGene.setText("Fitness:"+b.Fitness);
      MyUniverse.ViewCurrentBitGen.DrawBitGenome(b);
      MyUniverse.ViewGene2.DrawGenome(b);
      repaint();
      return(true);
    }
    if(Type==TYPEZOOM) {
      Type=TYPEMAIN;
      MyUniverse.App.EnableRun();
      SetSteps();
      MyUniverse.DrawWorld();
      return(true);
    }

    if(!MyUniverse.IsInit) {
      b=MyUniverse.CurrentBiom;
      MyUniverse.SetCurrentBiom(SelectedGrid-1);
      b.Live(this);
      MyUniverse.CurrentBiom.Live(this);
    }
    else if(SelectedGrid!=1) {
      b=MyUniverse.SelectedBiom;
      MyUniverse.SetSelectedBiom(SelectedGrid-1);
      MyUniverse.ViewCurrentBitGen.DrawBitGenome(MyUniverse.SelectedBiom);
      b.Live(this);
      MyUniverse.SelectedBiom.Live(this);
    }
    repaint();
    MyUniverse.DispStats();
    return(true);
  }

  /** Set # of places */
  public void SetGrid(int n) {
    switch(n) {
      case 1:
        NbHorizPlace=2;
        NbVerticPlace=1;
      break;
      case 2:
        NbHorizPlace=3;
        NbVerticPlace=3;
      break;
      case 4:
        NbHorizPlace=10;
        NbVerticPlace=10;
      break;
      case 5:
        NbHorizPlace=22;
        NbVerticPlace=22;
      break;
      case 6:
        NbHorizPlace=30;
        NbVerticPlace=30;
      break;
      case 7:
        NbHorizPlace=40;
        NbVerticPlace=40;
      break;
      case 3:
      default:
        n=3;
        NbHorizPlace=5;
        NbVerticPlace=5;
    }
    VisiblePop=NbHorizPlace*NbHorizPlace;
    SetSteps();
    CurrentGrid=n;
  }

  /** Returns current grid */
  public int GetGrid() {
    return(CurrentGrid);
  }
  /** Returns current visible pop */
  public int GetVisiblePop() {
    return(VisiblePop);
  }
}


