//***************************************************************************
//  JPGraphTools 1.0
//    10/1/2000
//    Diverse Graphical tools.
//
//    Jean-Philippe Rennard
//    alife@rennard.org
//    http://www.rennard.org/alife
//***************************************************************************
package Gav;

import java.awt.*;
//***************************************************************************
/** class JPGraphTools : Various Graphic functions **/
class JPGraphTools
{
    /** Type of Representation **/
    static final int REPRBASE=0;
    static final int REPRANISOTROPIC=1;
    static final int REPRISOTROPIC=2;
    /** Current Reprensentation **/
    int ReprCrt;
    /** Logical size (theorical size for calculation) **/
    float lWidth,lHeight;
    /** Pixel size according to logical and real sizes **/
    float pWidth,pHeight;
    /** Current Canvas size **/
    int MaxX,MaxY;
    /** Origin Offsets **/
    int OrigOffsetX,OrigOffsetY;
    /** default constructor. Init size to 0 **/
    public JPGraphTools() {
      ReprCrt=REPRBASE;
      Resize(0,0);
    }
    /** Constructor. Receives size of the working Canvas **/
    public JPGraphTools(int mx, int my,int repr) {
      SetRepresentation(repr);
      Resize(mx,my);
    }
    /** set current origin offsets **/
    public void SetOrigOffset(int x,int y) {
      OrigOffsetX=x;
      OrigOffsetY=y;
    }
    /**set current working size **/
    public void Resize(int mx,int my) {
      MaxX=mx-1;
      MaxY=my-1;
      OrigOffsetX=0;
      OrigOffsetY=MaxY;
      SetLogicSize(MaxX,MaxY);
    }
    /**set current working and logical size **/
    public void Resize(int mx,int my,float lx,float ly) {
      MaxX=mx-1;
      MaxY=my-1;
      OrigOffsetX=0;
      OrigOffsetY=MaxY;
      SetLogicSize(lx,ly);
    }
    /** set Representation **/
    public void SetRepresentation(int r) {
      switch(r) {
        case REPRANISOTROPIC:
        case REPRISOTROPIC:
          ReprCrt=r;
          break;
        default :
          ReprCrt=REPRBASE;
      }
    }
    /** Get Representation **/
    public int GetRepresentation() {
      return(ReprCrt);
    }
    /** set logical size **/
    public void SetLogicSize(float w,float h) {
      lWidth=w;
      lHeight=h;
      switch(ReprCrt) {
        case REPRANISOTROPIC:
          pWidth=lWidth/MaxX;
          pHeight=lHeight/MaxY;
          break;
        case REPRISOTROPIC:
          pWidth=pHeight=Math.max(lWidth/MaxX,lHeight/MaxY);
          break;
        default:
          pWidth=1;
          pHeight=1;
      }
    }


    /** Logical to system display conversion
    Return : x=x, y=MaxY-y */
    public int LToRX(float x_) {
      switch(ReprCrt) {
        case REPRBASE:
          if(x_>MaxX)
            return(MaxX);
          return(Math.round(x_)+OrigOffsetX);
        case REPRANISOTROPIC:
          return(Math.round((x_/pWidth))+OrigOffsetX);
        case REPRISOTROPIC:
          return(Math.round(x_/pWidth)+OrigOffsetX);
      }
      return(0);
    }
    public int LToRY(float y_) {
      switch(ReprCrt) {
        case REPRBASE:
          if(y_>MaxY)
            return(0);
          return(OrigOffsetY-Math.round(y_));
        case REPRANISOTROPIC:
          return(OrigOffsetY-Math.round(y_/pHeight));
        case REPRISOTROPIC:
          return(OrigOffsetY-Math.round(y_/pHeight));
      }
        return(0);
    }

    /** System display to Logical conversion */
    public float RToLX(int x_) {
      switch(ReprCrt) {
        case REPRBASE:
          if(x_>MaxX+OrigOffsetX)
            return(MaxX);
          return(x_-OrigOffsetX);
        case REPRANISOTROPIC:
        case REPRISOTROPIC:
          return((x_-OrigOffsetX)*pWidth);
      }
      return(0);
    }
    public float RToLY(int y_) {
      switch(ReprCrt) {
        case REPRBASE:
          if(y_>MaxY)
            return(0);
          return(OrigOffsetY-MaxY-y_);
        case REPRANISOTROPIC:
          return( (OrigOffsetY-y_)*pHeight );
        case REPRISOTROPIC:
          return( (OrigOffsetY-y_)*pHeight );
      }
        return(0);
    }

} // JPGraphFunct
//***************************************************************************
/** PosReal : Implements Position in real (float) **/
class PosReal
{
  int NbDim;
  float x;
  float y;
  float z;
  /** Constructor 2D par dfaut**/
  public PosReal() {
    x=0;
    y=0;
    NbDim=2;
  }
  /** Constructor 2D **/
  public PosReal(float x_,float y_) {
    x=x_;
    y=y_;
    NbDim=2;
  }
  /** Constructor 3D **/
  public PosReal(float x_,float y_, float z_) {
    x=x_;
    y=y_;
    z=z_;
    NbDim=3;
  }
  /** Converts PosReal in java.awt.Point **/
   public void PosRealToPoint(Point pt) {
      pt.x=(int)Math.round(x);
      pt.y=(int)Math.round(y);
  }

} // PosReal
//***************************************************************************
